import logging
import asyncio
import zlib
import json
import time
import sys
import websockets
from .http_client import HTTP
from .utils import exceptions
from .utils import API
from .utils import Logging

log = logging.getLogger(__name__)
log.setLevel(Logging.gateway.LEVEL)
logging.basicConfig(format=Logging.FORMAT)

class WebSocket(websockets.client.WebSocketClientProtocol):
    OPCODE = {
        "DISPATCH"          : 0,
        "HEARTBEAT"         : 1,
        "IDENTIFY"          : 2,
        "PRESENCE"          : 3, 
        "VOICE_STATE"       : 4,
        "VOICE_PING"        : 5,
        "RESUME"            : 6,
        "RECONNECT"         : 7,  
        "REQUEST_MEMBERS"   : 8,    
        "INVALIDATE_SESSION": 9, 
        "HELLO"             : 10,
        "HEARTBEAT_ACK"     : 11,
        "GUILD_SYNC"        : 12,
    }

    class events:
        async def on_ready():
            pass

    def __init__(self, token, loop=None):
        self._zlib = zlib.decompressobj()
        self._buffer = bytearray()
        self.sequence = None
        self.heartbeat_interval = 0
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self.token = token
        self.recv_timeout = 0.1
        self.ZLIB_SUFFIX = b'\x00\x00\xff\xff'

        self.http = HTTP(token=self.token,loop=self.loop)
        self.gateway()

    def gateway(self):
        self.loop.run_until_complete(self.get_gateway())

    async def get_gateway(self):
        self.BASE = await self.http.get("/gateway")
        self.BASE = json.loads(self.BASE)["url"]

        log.info("Received gateway url")

        await self.connect()
        await self.identify()

        await self.send_heartbeat()

    async def connect(self):
        self.socket = await websockets.connect(self.BASE + f"/?v={API.LEVEL}&encoding=json")
        log.info("Connected to Discord API")

        discord_hello = await self.recv_data()
        self.heartbeat_interval = discord_hello["d"]["heartbeat_interval"]

    async def close(self):
        await self.socket.close()
        self.loop.stop()
        log.info("Stopped client")

    async def send_heartbeat(self):
        self.last_heartbeat = time.time()
        heartbeat_data = self.sequence
        data = self.data_obj(self.OPCODE["HEARTBEAT"], heartbeat_data)
        await self.send_data(data)

    async def identify(self):
        identify_data = {
            "token": self.token,
                "properties": {
                    "$os": sys.platform,
                    "$browser": "discordo",
                    "$device": "discordo"
                },
                "compress": False,
                "large_threshold": 250,
        }
        data = self.data_obj(self.OPCODE["IDENTIFY"], identify_data)
        await self.send_data(data)

    def data_obj(self, op, d, t=""):
        data = {"op": op, "d": d, "t": t, "s": self.sequence}
        return data

    async def send_data(self, data):
        json_data = json.dumps(data)
        
        try:
            await self.socket.send(json_data)
        except Exception as error:
            self.handle_websocket_errors(error)

        log.debug("send: Sequence: " + str(data["s"]))
        log.info("send: Opcode: " + str(data["op"]) + " (" + list(self.OPCODE.keys())[data["op"]] + ")")

        if data["d"] != None and "{" in json.dumps(data["d"], indent=2):
            log.debug("send: Data: \n" + json.dumps(data["d"], indent=2) + "\n")
        elif data["d"] != None:
            log.debug("send: Data: " + json.dumps(data["d"], indent=2) + "\n")     

        self.update_sequence()

    def handle_websocket_errors(self, error):
        if str(error) != "":
            log.error(error)

            if "Authentication failed" in str(error):
                raise exceptions.AuthenticationFailed("Authentication failed.")

            if "connection closed abnormally" in str(error):
                raise exceptions.UnknownError("Connection closed.")

    async def recv_data(self):
        try:
            json_data = await asyncio.wait_for(self.socket.recv(), self.recv_timeout)
        except Exception as error:
            return self.handle_websocket_errors(error)
                 
        data = json.loads(json_data)
        self.update_sequence(data)

        log.debug("recv: Gateway Event Name: " + str(data["t"]))
        log.debug("recv: Sequence: " + str(data["s"]))
        log.info("recv: Opcode: " + str(data["op"]) + " (" + list(self.OPCODE.keys())[data["op"]] + ")")

        if data["d"] != None and "{" in json.dumps(data["d"], indent=2):
            log.debug("recv: Data: \n" + json.dumps(data["d"], indent=2) + "\n")
        elif data["d"] != None:
            log.debug("recv: Data: " + json.dumps(data["d"], indent=2) + "\n")

        return data

    async def event_loop(self):
        data = await self.recv_data()
        if data != None:
            if data["t"] == "READY":
                log.info("Client ready")
                await self.events.on_ready(data)

    def update_sequence(self, received_data=None):
        if received_data != None:
            if received_data["s"] != None:
                self.sequence = received_data["s"]
                if self.sequence != received_data["s"] and self.sequence != None:
                    log.error("Sequence missmatch")

        else:
            if self.sequence != None:
                self.sequence += 1

    def decode_data(self, data):
        pass

    def on_websocket_message(msg):
        print(msg)
        self._buffer.extend(msg)

        if len(msg) < 4 or msg[-4:] != self.ZLIB_SUFFIX:
            return

        msg = self._zlib.decompress(buffer).decode('utf-8')
        self._buffer = bytearray()